/* GSI - general sound interface  [-*- C++ -*-]
 *
 * gsi.h
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 */

/* You can use, redistribute and/or modify this software under the terms of
 * the CGPL version 0.4 or later (see the file LICENSE.CGPL for details)
 */


#ifndef _gsi_h
#define _gsi_h

//Shared
#include <gsi/config.h>
#include <gsi/gsi_commands.h>

#ifndef USE_NAS
#  include <gsi/gsi_server_version.h>
#  define GSI_INTERFACE_VERSION GSI_SERVER_VERSION
#endif
#ifdef USE_NAS
#  include <gsi/gsi_nas_version.h>
#  define GSI_INTERFACE_VERSION NAS_INTERFACE_VERSION
#endif



#ifdef _GSI_INTERFACE_
/*---------------------------------------------------------------------------*
 *	Global declarations for the interface file
 *---------------------------------------------------------------------------*/
//#include <gsi/config.h>
//char *gsi_version;
//      A string to use for program startup etc

int  gsi_init ();
//      This initialises the GSI sound system (initialises various things,
//      such as the mixer settings when using GSI's gsi_server, and connects
//      to the NAS server when using the NAS wrapper) and is to be called
//      before using any other GSI function.

void gsi_close ();

int  gsi_init_pcm (int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile);
//      initialises the PCM system: sounddir and soundfile may be NULL or zero
//      length strings. In that case they are not set and no sounds will be
//      loaded.

int  gsi_set_midi_devices (int midichannels, int synthchannels);

int  gsi_init_synth (int type);

int  gsi_init_cd (int type);

void gsi_flush ();
//      send all cached commands to the sound server.
//      Also checks  for  callbacks  and  errors and calls the relevant
//      callback fuctions.

//Add gsi_hard_flush(); ?? sends all data even if write will block
//or gsi_sync() ??

void gsi_play_sound (int slot, int command);
//      play/kill sounds using dynamically allocated dsp channels (multiple
//      times the same sound is allowed).
//      Examples:
//          play_sound(k, SND_PLAY)  = play using new channel
//          play_sound(k, SND_REPEAT)  = ditto, repeat the sound if finished
//          play_sound(k, SND_OVERWRITE)  = overwrite LAST started k sound
//          play_sound(k, SND_KILL)  = kill LAST started k sound

//CHANGE ch/slot order ? (int slot, int channel, int command, ...)

void gsi_play_channel_sound (int channel, int slot, int command);
//      play a sound on a channel, other than that, equivalent to play_sound(.)

void gsi_play_sound_flush (int slot, int command);
//      same as play_sound but with immediate gsi_flush() appended.


//CHANGE ch/slot order ? (int slot, int channel, int command, ...)

//Special compound playing commands, use ch=-1 for stack playing
//      Note that speeds are the distance travelled in 1/32 s (automatically
//      recalulated for the given fragment size of the DSP, as positions are
//      adjusted each fragment)
void gsi_play_sound_3d (int ch, int slot, int command, int x, int y, int z, int dx, int dy, int dz, int doppler_on);

void gsi_play_sound_stereo (int ch, int slot, int command, int x, int dx, int doppler_on);

void gsi_play_sound_pan (int ch, int slot, int command, int pan);
//      pan =-256 = full left, 256=full right

void gsi_kill_sounds ();
//      stops all sounds on the stack and on channels

int  gsi_get_channel_status (int ch);
//      status of channel ch, is there a sound playing on it?


// Midi loading (note that all these commands stop any playing song first):
int  gsi_load_song_file (char *filename);
//rename: gsi_send_song_data or gsi_load_song_data ??

int  gsi_load_song_data (int data_len, unsigned char *data);

int  gsi_load_song_track (int trackno, int data_len, unsigned char *data);
//      For music making programs in particular

int  gsi_set_song_controller (int midi_ctl, int midi_ch);
//      To set the pan, volume etc for each of the midi channels

// Midi playing:
int  gsi_play_song (int loopno);
//      Starts playing the currently loaded song <loopno> times. If <loopno>=0
//      it will continue to play until it is stopped or a new song is loaded.

int  gsi_stop_song ();
//      Stops playing the currently loaded song

int  gsi_continue_song ();
//      Continues playing the currently loaded song (if given before a
//      gsi_play_song, this is ignored)

//Individual track playing ??? gsi_stop_track(int track)


// Custom patch loading:
int  gsi_load_patch (int pgm, char *patchname);
int  gsi_load_patch_data (int pgm, int patchlen, unsigned char *patchdata);
// type of device? GUS/OPL3/other

void gsi_send_commands (int argc, ...);
//      send any commands to the soundserver (in byte chunks; 16bit and larger
//      must be sent LSB first) { change to MSB? }
//	    eg. gsi_send_commands(5, SND_SET_X | SND_CURRENT, x &255, x >>8,
//          SND_SET_DX | SND_CURRENT, dx);

void gsi_grab_pcm ();
//      grabs the DSP and sets the speed, channels etc. as set by earlier commands.

void gsi_grab_synth ();
//      grabs the sequencer

void gsi_grab_cd ();
//      grabs the cd

void gsi_release_pcm ();
//      releases the DSP (to share it with other programs)

void gsi_release_synth ();
//      releases the sequencer

void gsi_release_cd ();
//      releases the cd


// Do vol_L/vol_R ? return value = L | (R<<8), set : (int volL, int volR)
void gsi_set_master_volume (int vol);
//      If there's no master volume, emulate it with an internal volume and
//      adjust all others (SYNTH, PCM, CD) accordingly

void gsi_set_pcm_volume (int vol);
void gsi_set_synth_volume (int vol);
void gsi_set_cd_volume (int vol);
int  gsi_get_master_volume ();
int  gsi_get_pcm_voglume ();
int  gsi_get_synth_volume ();
int  gsi_get_cd_volume ();

// Unimplemented:
void gsi_cd_init ();
void gsi_cd_stop ();
void gsi_cd_mount ();
void gsi_cd_unmount ();
void gsi_cd_eject_disk ();
//void gsi_cd_insert_disk ();
void gsi_cd_start_play ();
void gsi_cd_restart_track ();
void gsi_cd_next_track ();
void gsi_cd_previous_track ();
void gsi_change_cd (int cdno); /* For CD changers */

// Low level
void _gsi_send_byte (int k);
void _gsi_send_int16 (int k);
void _gsi_send_int32 (int k);
void _gsi_send_string (char *s);
void _gsi_send_block (unsigned char /*byte*/ *data, int len);

// Calback functions:
void *gsi_set_channel_callback (int ch, void (*cbf)(int ch));
void gsi_clear_channel_callback (int ch);
// Following not needed: channel -1 is the song in above functions
//void *gsi_set_song_callback
//void gsi_clear_song_callback

void *gsi_set_error_callback (void (*cbf)(int error));
void gsi_clear_error_callback ();

int  gsi_wait_for_callback (int musec);
//    waits using select up to time micro seconds (i.e. max 2147 seconds)

int  gsi_callback_fd ();
//    returns the file descriptor used for callback information so you can use
//    it in a select together with other descriptors


/////////////////////////////////////////////
// Dynamic allocation ? This is a prototype:
int  gsi_get_free_slot();
//    returns -1 when all slots are used, 0-255 otherwise.
int  gsi_get_free_channel();
//    returns -1 when all channels are playing sounds, 0-255 otherwise
//SYNC necesssary for both (i.e. slows the program .02 - 0.4 s for most unixes as tasks must be switched twice), or use user applied locks:
int  gsi_lock_channel(int ch);
int  gsi_lock_slot(int slot);
// Then wait for callbacks to set the flags definitely
// ?? gsi_get_channel_lock(int ch), gsi_get_slot_lock(slot): true if locked
/////////////////////////////////////////////

#else
/*---------------------------------------------------------------------------*
 *	External (publicly available)
 *---------------------------------------------------------------------------*/
extern char *gsi_version;

extern int  sound_on;

extern int  gsi_init ();
extern void gsi_close ();
extern int  gsi_init_pcm (int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile);
extern int  gsi_init_synth (int type);

extern void gsi_play_sound (int slot, int command);
extern void gsi_play_channel_sound (int channel, int slot, int command);


extern void gsi_play_sound_3d (int ch, int slot, int command, int x, int y, int z, int dx, int dy, int dz, int doppler_on);
extern void gsi_play_sound_stereo (int ch, int slot, int command, int x, int dx, int doppler_on);
extern void gsi_play_sound_pan (int ch, int slot, int command, int pan);



extern void gsi_flush ();
extern void gsi_play_sound_flush (int slot, int command);

extern int  gsi_get_channel_status (int ch);

#define gsi_kill_sound(slot)       gsi_play_sound((slot), SND_KILL)
#define gsi_kill_channel_sound(ch) gsi_send_commands(2, SND_CHANNEL_KILL, (ch))

extern void gsi_kill_sounds ();


extern int  gsi_play_song (int loopno);
extern int  gsi_stop_song ();
extern int  gsi_continue_song ();

extern void gsi_grab_pcm ();
extern void gsi_grab_synth ();
extern void gsi_grab_cd ();

extern void gsi_release_pcm ();
extern void gsi_release_synth ();
extern void gsi_release_cd ();

extern void gsi_set_master_volume (int vol);
extern void gsi_set_pcm_volume (int vol);
extern void gsi_set_synth_volume (int vol);
extern void gsi_set_cd_volume (int vol);
extern int  gsi_get_master_volume ();
extern int  gsi_get_pcm_volume ();
extern int  gsi_get_synth_volume ();
extern int  gsi_get_cd_volume ();

extern void gsi_send_sample (int slot, int speed, int type, int channels, int size, unsigned char *data);
extern int  gsi_load_file (int slot, char *filename);

extern int  gsi_load_song_file (char *filename);
extern int  gsi_load_song_data (int data_len, unsigned char *data);
extern int  gsi_load_song_track (int trackno, int data_len, unsigned char *data);


extern void gsi_send_commands (int argc, ...);
//This was wrong:
//#define gsi_send_byte         _gsi_send_byte
//#define gsi_send_int16        _gsi_send_int16
//#define gsi_send_int32        _gsi_send_int32
//#define gsi_send_string       _gsi_send_string
//#define gsi_send_block        _gsi_send_block 
//as we need to flush before writing to the pipe!
//IMPLEMENT NEW ONES!

//Unimplemented:
extern void gsi_cd_init ();
extern void gsi_cd_stop ();
extern void gsi_cd_mount ();
extern void gsi_cd_unmount ();
extern void gsi_cd_eject_disk ();
extern void gsi_cd_start_play ();
extern void gsi_cd_restart_track ();
extern void gsi_cd_next_track ();
extern void gsi_cd_previous_track ();
extern void gsi_change_cd (int cdno);


extern void *gsi_set_channel_callback (int ch, void (*cbf)(int ch));
extern void gsi_clear_channel_callback (int ch);

extern void *gsi_set_error_callback (void (*cbf)(int error));
extern void gsi_clear_error_callback ();

extern int  gsi_wait_for_callback (int musec);
extern int  gsi_callback_fd ();

#endif /* INTERFACE / ! INTERFACE */


#endif /* _gsi_h */
