/* GSI - general sound interface  [-*- C++ -*-]
 * 
 * gsi_server.h
 *
 * Copyright (C) W.H.Scholten 1998
 *
 */

#ifndef _GSI_SERVER_H
#define _GSI_SERVER_H

#include <gsi/mytypes.h>

#define min(x,y) ((x)>(y) ? (y): (x))
#define max(x,y) ((x)<(y) ? (y): (x))


/* Make these maxima dynamic? */
#define MAX_SOUNDS           256 /* Maximum number of sounds loaded (slots), this can have any value you want (note that it uses n*4 bytes for the 'playing' table), actually the interface sends 1 byte, so could use byte playing[..] and restrict to 256 always) */
/* perhaps use a 1-2 byte scheme to use 0-127, 128-255 +2nd byte for 128+32768 possible samples */

// These two should go to a separate config file:
#define MAX_STATIC_CHANNELS   64 /* Maximum number of simultaneous sounds on channels (the maximum of this value is 256) */
#define MAX_DYNAMIC_CHANNELS  32 /* Maximum number of simultaneous sounds dynamically allocated on a stack (the maximum of this value is only restricted by memory) */
#define MAX_CHANNELS        (MAX_STATIC_CHANNELS + MAX_DYNAMIC_CHANNELS)

#define MAX_STATUS_QUEUE_LEN 256

/* Defaults are in comments */

typedef struct client_data {
	int  dsp_request; /* type of connection we request: AUTO_DSP, GRAB_DSP, */
	int  new_dsp_speed /*=8000*/;
	int  new_dsp_log2_frag /*=8*/;
	int  new_dsp_channels /*=1*/;
	int  new_dsp_type /*=GSI_8BIT*/;
int  paused;

	int  autorelease_dsp /*False*/;
	/* If true, the DSP is released when no sounds are playing, it will be
	 * grabbed again when a 'play sound' command is received */
	int  autoset_dsp /*=False*/;
	/* If true, the DSP settings are changed whenever a sound is played after
	 * a period of silence, to the rate/bytes/channels of that sound */
	int  do_callback /*=False*/;
	/* If true, callbacks will be sent whenever a sound is started or killed
	 * on a channel (not the stack) */

	/* ** Decoding the current command from bytes read from the pipe (async mode) */
	byte command_size /* 0 */;       // 0= no command processed
	byte command;            // if command_size>0 this is the current command
	byte command_data[32];   // all further bytes from a command
	// Note: commands that send a lot of data (e.g. SND_LOAD_SOUND_DATA) will
	// not be handled in async mode.

	// These two are also set by sound modify commands:
	int  last_sound_stack /* -1 */;   // stack no of last started sound
	int  last_sound_channel /* -1 */;
	/* channel of last command (for immediately following update commands),
	 * -1 if none (done playing or a kill command issued on the channel) */
	byte channel_used [MAX_STATIC_CHANNELS];
	/* simple linear list of which channels are playing */


	char sounds_dir[256];
	int  num_sounds /*=0*/;       /* = 1+ highest number of sound slots used */


	/* ** Sounds loaded into memory */
	int  sounds_loaded;
int  sounds_file_loaded /*=False*/;

	/* have the sounds from the sound list been loaded? */
	byte sound_from_list [MAX_SOUNDS]; /* incorporate into sound_flags ? */
	/* is it a sound from the list or separately loaded? useful? */
	char *sound_filename [MAX_SOUNDS]; /* file from which the slot was loaded */
	int8 *sound_buffer [MAX_SOUNDS]; //the actual sound data
	int  sound_size [MAX_SOUNDS];
	/* bytes, >>should become samples<< ? (16/8 bit..) */
	int  sound_speed [MAX_SOUNDS];
	byte sound_type [MAX_SOUNDS];     /* 8/16bit, 8 bit mu-law */
	byte sound_channels [MAX_SOUNDS]; /* mono, stereo */
	int  sound_conversion [MAX_SOUNDS];
	int  sound_conversion2 [MAX_SOUNDS]; /* recalculated at the moment */
	/* conversion from sound to dsp speed (will be replaced by RL algorithm) */


	/* ** Data to control sound effects, speed conversion etc */
	int  base_width /*=256*/; /* 0-256, 0=mono, 256=full stereo */
	short sound_pan [MAX_CHANNELS];
	int  old_dist [MAX_CHANNELS]; /* For doppler */
	short sound_volume [MAX_CHANNELS]; /* saved value, used when a sound doen't move to save recalculation */

	int  sound_x [MAX_CHANNELS];
	int  sound_z [MAX_CHANNELS];
	int  sound_y [MAX_CHANNELS];

	short sound_dx [MAX_CHANNELS];
	short sound_dz [MAX_CHANNELS];
	short sound_dy [MAX_CHANNELS];

	// Time at which the position was last set/adjusted with SND_SET_X etc, and
	// the current_time (set to zero at init)
	// This is measured in 1/32 s, 24 bit int, 8 bit fraction

	int  last_position_time [MAX_CHANNELS];
	int  current_time /*=0*/;

	short sound_echo_dt [MAX_CHANNELS]; // dt*32
	byte sound_echo_fvol [MAX_CHANNELS];
	/* fractional volume change, should be 1/(8R^3) where R is distance to the
	 * bounce point => log2: -3 -3* 2log(R) ? => dvol? */

	//OBSOLETE: short  sound_doppler [MAX_CHANNELS]; absolute value of speed change, actually (r'-r)/dt, i.e. when going in a circle, doppler=0 !

	int  observer_x /*= 0*/;
	int  observer_z /*= 0*/;
	int  observer_y /*= 0*/;
	int  observer_angle /*= 0*/; //fi
	int  observer_theta /*= 0*/;
	/* For observer directioning: observer_look_fi/observer_look_theta */

	int  observer_dx /*= 0*/;
	int  observer_dz /*= 0*/;
	int  observer_dy /*= 0*/;

	/* ** Current state of sounds playing */
	int  playing [MAX_CHANNELS];   /* Sound file numbers currently being played */
	int  position [MAX_CHANNELS];  /* Current position in file for each sound */
	int  sound_flags [MAX_CHANNELS];

	int  playnum /*= 0*/;              /* Number of sounds currently being played on the stack */
	int  channel_playnum /*= 0*/;      /* Number of sounds currently being played on channels */

	int  max_stack /*=MAX_DYNAMIC_CHANNELS*/; /* max number of simultaneous sounds via the stack, reduce with SND_SET_MAX_AUTO_CHANNELS to limit the maximum CPU use */

	int  mach1 /*=256<<8*/; /* the speed of sound = 256.0 :-) */

	// For callbacks:
	byte status_queue[MAX_STATUS_QUEUE_LEN];
	int  status_queue_len;

    int  socket_fd; /* File descriptor for the socket */
} client_data;


#endif /* _GSI_SERVER_H*/
