/* GSI - general sound interface
 *
 * gsi_midi_play_synth.c
 *
 * Copyright (C) W.H.Scholten 1998
 *
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <gsi/config.h>
#include <gsi/mytypes.h>
#include <gsi/debug.h>
#include "midi.h"

#ifdef __linux__
#  include <sys/soundcard.h>
#  include <linux/awe_voice.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
//#  include <machine/awe_voice.h> /* ?? */
#endif

#ifdef USE_OSS /* This code is currrently OSS only */


/*__________________EXTERNALS___________________*/


/*__________________VARIABELEN___________________*/
SEQ_DEFINEBUF(1024);
/*extern int _seqbufptr;*/


struct voice_state **voices; /* For each sound card separate */

/* Midi state: */
struct channel_state channel[16];
int  note_vel[16][128];

/*---------------------------------------------------------------------------*
 *	universal sysex reset GM synth controllers (not used)
 *---------------------------------------------------------------------------*/
unsigned char gm_reset[5] = {
	0x7e, 0x7f, 0x09, 0x01, 0xf7
};


/*__________________PROGRAMMA___________________*/

/*---------------------------------------------------------------------------*
 *	Reset the sequencer and external midi devices
 *---------------------------------------------------------------------------*/
void seq_reset(/*int seq_fd*/)
{
	int i, j;

	DEB1(fprt(stderr, "__GSI: seq_reset\n"));

	_seqbufptr = ticks = 0;
	ioctl(seqfd, SNDCTL_SEQ_RESET);

	for (i = 0; i < 16; i++) {
		if (IS_MIDI(i)) {
			seq_control(channel_output[i],i,0,0);
			seq_control(channel_output[i],i,32,0);
		}
		seq_set_patch(i, 0);
		for (j = 0; j < 128; j++)
			note_vel[i][j] = 0;
		channel[i].bender = 8192;
		channel[i].bender_range = 2;
		channel[i].controller[CTL_PAN] = 64;
		channel[i].controller[CTL_SUSTAIN] = 0;
	}


	for (j=0; j<nrsynths; j++) {
		int  synth_type=synth_types[channel_output[j]];
#ifdef USE_OPL3_SYNTH
		if (synth_type==SYNTH_HW_OPL) {
			if (use_opl3)
				ioctl(seqfd, SNDCTL_FM_4OP_ENABLE, &j);
		}
#endif
#if defined (USE_GUS_SYNTH) || defined(USE_OPL3_SYNTH)
		if (synth_type==SYNTH_HW_OPL || synth_type==SYNTH_HW_GUS) {
			for (i = 0; i < card_info[j].nr_voices; i++) {
				SEQ_CONTROL(j, i, SEQ_VOLMODE, VOL_METHOD_LINEAR);
//				SEQ_CONTROL(j, i, SEQ_VOLMODE, VOL_METHOD_ADAGIO);

				if (voices[0][i].note)
					SEQ_STOP_NOTE(j, i, voices[0][i].note, 64);
				voices[0][i].available = voices[0][i].timestamp = -1;
			}
		}
#endif
#ifdef USE_AWE_SYNTH
		if (synth_type==SYNTH_HW_AWE) {
			AWE_SET_CHANNEL_MODE(j, 1);
			AWE_DRUM_CHANNELS(j, perc_chn);
			AWE_TERMINATE_ALL(j);
			for (i = 0; i < card_info[j].nr_voices; i++) {
				voices[0][i].available = voices[0][i].timestamp = -1;
			}
		}
#endif
	}
	SEQ_DUMPBUF();
	DEB1(fprt(stderr, "__GSI: seq_reset done\n"));

}


/*---------------------------------------------------------------------------*
 *	set a patch for channel chn to pgm
 *---------------------------------------------------------------------------*/
int seq_set_patch(int chn, int pgm)
{
	DEB1(fprt(stderr, "__GSI: seq_set_patch\n"));

	if (force_chn_pgm[chn])
		pgm = force_chn_pgm[chn] - 1;

	switch(synth_types[channel_output[chn]]) {
	case MIDI_HW:
		SEQ_MIDIOUT(channel_output[chn], MIDI_PGM_CHANGE + chn);
		SEQ_MIDIOUT(channel_output[chn], pgm);
		break;
	case SYNTH_HW_OPL:
#ifdef USE_OPL_SYNTH
		if (!fmloaded[pgm])
			if (IS_PERC(chn))
				return -1;
#endif
		break;
	case SYNTH_HW_GUS:
#ifdef USE_GUS_SYNTH
		if (patchloaded[pgm] != 1) {
			if (IS_PERC(chn))
				return -1;
			else {
				/* must replace any missing one, with knowledge of which one
				 * can be missed at the moment. For now, we simply replace the
				 * missing one with the first loaded gus program (make a
				 * similarity table ? or the closest in pgm number ?) */
				for (pgm = 0; pgm<256; pgm++) {
					if (patchloaded[pgm] == 1) break;
				}
			}
		}
#endif
		break;
	case SYNTH_HW_AWE32:
#ifdef USE_AWE_SYNTH
		SEQ_SET_PATCH(channel_output[chn], chn, pgm);
#endif
		break;
	case SYNTH_HW_PCM_EMULATION:
		break; /* not yet implemented */
	}

	return (channel[chn].program = pgm);
}


void seq_stop_note(int dev, int chn, int note, int vel)
{
	int i;

	DEB1(fprt(stderr, "__GSI: seq_stop_note dev=%d, chn=%d, note=%d, vel=%d\n", dev, chn, note, vel));
	note_vel[chn][note] = 0;

	switch(synth_types[channel_output[chn]]) {
	case MIDI_HW:
		SEQ_MIDIOUT(dev, MIDI_NOTEOFF + chn);
		SEQ_MIDIOUT(dev, note);
		SEQ_MIDIOUT(dev, vel);
		break;
	case SYNTH_HW_AWE32:
		SEQ_STOP_NOTE(dev, chn, note, vel);
		break;
	default:
		for (i = 0; i < card_info[dev].nr_voices; i++)
			if (voices[dev][i].channel == chn && voices[dev][i].note == note) {
				voices[dev][i].available = 1;
				voices[dev][i].timestamp /= 2;
				if (!channel[chn].controller[CTL_SUSTAIN] && !IS_PERC(chn))
					SEQ_STOP_NOTE(dev, i, note, vel);
			}
	}
}


void seq_key_pressure(int dev, int chn, int note, int vel)
{
	int i;

	DEB1(fprt(stderr, "__GSI: seq_key_pressure dev=%d, chn=%d, note=%d, vel=%d\n", dev, chn, note, vel));

	switch(synth_types[channel_output[chn]]) {
	case MIDI_HW:
		SEQ_MIDIOUT(dev, MIDI_KEY_PRESSURE + chn);
		SEQ_MIDIOUT(dev, note);
		SEQ_MIDIOUT(dev, vel);
		break;
	case SYNTH_HW_AWE32:
		AWE_KEY_PRESSURE(dev, chn, note, vel);
		break;
	default:
		for (i = 0; i < card_info[dev].nr_voices; i++)
			if (voices[dev][i].channel == chn && voices[dev][i].note == note)
				SEQ_KEY_PRESSURE(dev, i, note, vel);
	}
}


/* returns oldest voice (first started) */
int new_voice(int dev, int chn)
{
	int  i, oldest, last;

	DEB1(fprt(stderr, "__GSI: new_voice dev=%d, chn=%d\n", dev, chn));
	if (IS_FM(chn) && fmloaded[channel[chn].program] == OPL3_PATCH)
		last = 6;		/* A 4-op voice can only use the first six voices */
	else
		last = card_info[dev].nr_voices;

	for (i = oldest = 0; i < last; i++) {
		if (voices[dev][i].timestamp < voices[dev][oldest].timestamp)
			oldest = i;
	}

	return oldest;
}


void seq_start_note(int dev, int chn, int note, int vel)
{
	int  c;
	int  voice;

	DEB1(fprt(stderr, "__GSI: seq_start_note dev=%d, chn=%d, note=%d, vel=%d\n", dev, chn, note, vel));
	note_vel[chn][note] = vel;

	switch(synth_types[channel_output[chn]]) {
	case MIDI_HW:
		SEQ_MIDIOUT(dev, MIDI_NOTEON + chn);
		SEQ_MIDIOUT(dev, note);
		SEQ_MIDIOUT(dev, vel);
		break;
	case SYNTH_HW_AWE32:
		if (vel == 0)
			seq_stop_note(dev, chn, note, 64);
		else
			SEQ_START_NOTE(dev, chn, note, vel);
		break;
	default:
		if (vel == 0)
			seq_stop_note(dev, chn, note, 64);
		else {
			voice = new_voice(dev, chn);
			SEQ_SET_PATCH(dev, voice, channel[chn].program);

		SEQ_CONTROL(dev, voice, CTRL_PITCH_BENDER_RANGE, (channel[chn].bender_range * 100));
		SEQ_BENDER(dev, voice, channel[chn].bender);
		SEQ_CONTROL(dev, voice, CTL_PAN, channel[chn].controller[CTL_PAN]);
		SEQ_START_NOTE(dev, voice, note, vel);
		voices[dev][voice].note = note;
		voices[dev][voice].channel = chn;
		voices[dev][voice].timestamp = ticks;
		voices[dev][voice].available = 0;

		/* Chorus handling */
		if ((c = channel[chn].controller[CTL_CHORUS_DEPTH] * 8)) {
			if (channel[chn].bender_range)
				c /= channel[chn].bender_range;
			voice = new_voice(dev, chn);
			SEQ_SET_PATCH(dev, voice, channel[chn].program);

			SEQ_CONTROL(dev, voice, CTRL_PITCH_BENDER_RANGE, (channel[chn].bender_range * 100));
			if (channel[chn].bender + c < 0x4000) {
				SEQ_BENDER(dev, voice, channel[chn].bender + c);
			} else {
				SEQ_BENDER(dev, voice, channel[chn].bender - c);
			}
			/* put chorus note on the left/right */
			c = channel[chn].controller[CTL_PAN];
#if 1 /* accentuate chorus direction */
			if (c < 64)
				c = 0;
			else if (c > 64)
				c = 127;
#endif
			SEQ_CONTROL(dev, voice, CTL_PAN, c);
			SEQ_START_NOTE(dev, voice, note, vel);
			voices[dev][voice].note = note;
			voices[dev][voice].channel = chn;

			/* chorus note lifespan */
			voices[dev][voice].timestamp = ticks / 2;
			voices[dev][voice].available = 0;
		}
	}
	}
}


/* REGIST_PARM_NUM_MSB / REGIST_PARM_NUM_LSB */
static int rpn_msb[16] = {
	127, 127, 127, 127, 127, 127, 127, 127,
	127, 127, 127, 127, 127, 127, 127, 127
};
static int rpn_lsb[16] = {
	127, 127, 127, 127, 127, 127, 127, 127,
	127, 127, 127, 127, 127, 127, 127, 127
};


void seq_control(int dev, int chn, int p1, int p2)
{
	int i;

	DEB1(fprt(stderr, "__GSI: seq_control dev=%d, chn=%d, controller parm1=%d, conroller parm2=%d\n", dev, chn, p1, p2));
	channel[chn].controller[p1] = p2;

	if (IS_MIDI(chn)) {
		SEQ_MIDIOUT(dev, MIDI_CTL_CHANGE + chn);
		SEQ_MIDIOUT(dev, p1);
		SEQ_MIDIOUT(dev, p2);
	}

	switch (p1) {
	case CTL_SUSTAIN:
		switch(synth_types[channel_output[chn]]) {
		case MIDI_HW:
			break;
		case SYNTH_HW_AWE32:
			SEQ_CONTROL(dev, chn, p1, p2);
			break;
		default:
			if (p1 == CTL_SUSTAIN && !p2) {
				for (i = 0; i < card_info[dev].nr_voices; i++)
					if (voices[dev][i].channel == chn && voices[dev][i].available) {
						SEQ_STOP_NOTE(dev, i, voices[dev][i].note, 64);
						voices[dev][i].available = 0;
					}
			}
		}
		break;
	case CTL_REGIST_PARM_NUM_MSB:
		rpn_msb[chn] = p2;
		break;
	case CTL_REGIST_PARM_NUM_LSB:
		rpn_lsb[chn] = p2;
		break;
	case CTL_DATA_ENTRY:
		if (rpn_msb[chn] == 0 && rpn_lsb[chn] == 0) {
			channel[chn].bender_range = p2;
			rpn_msb[chn] = rpn_lsb[chn] = 127;

			switch(synth_types[channel_output[chn]]) {
			case MIDI_HW:
				break;
			case SYNTH_HW_AWE32:
#if 0
				/* must find voice number .. */
				SEQ_CONTROL(dev, voice, CTRL_PITCH_BENDER_RANGE, (channel[chn].bender_range * 100));
#endif
				break;
			default:
				for (i = 0; i < card_info[dev].nr_voices; i++)
					SEQ_CONTROL(dev, i, CTRL_PITCH_BENDER_RANGE, p2 * 100);
			}
		}
		break;
	default:
		switch(synth_types[channel_output[chn]]) {
		case MIDI_HW:
			break;
		case SYNTH_HW_AWE32:
			SEQ_CONTROL(dev, chn, p1, p2);
			break;
		default:
			if ((p1 & 0xf0) == 0x00 || (p1 & 0xf0) == 0x50) {
				for (i = 0; i < card_info[dev].nr_voices; i++)
					if (voices[dev][i].channel == chn)
						SEQ_CONTROL(dev, i, p1, p2);
			}
			break;
		}
	}
	DEB1(fprt(stderr, "seq control done\n"));
}


void seq_chn_pressure(int dev, int chn, int vel)
{
	int  i;

	DEB1(fprt(stderr, "__GSI: seq_chn_pressure dev=%d, chn=%d, vel=%d\n", dev, chn, vel));
	channel[chn].pressure = vel;

	switch(synth_types[channel_output[chn]]) {
	case MIDI_HW:
		SEQ_MIDIOUT(dev, MIDI_CHN_PRESSURE + chn);
		SEQ_MIDIOUT(dev, vel);
	case SYNTH_HW_AWE32:
		AWE_CHN_PRESSURE(dev, chn, vel);
	default:
		for (i = 0; i < card_info[dev].nr_voices; i++)
			if (voices[dev][i].channel == chn)
				SEQ_KEY_PRESSURE(dev, i, voices[dev][i].note, vel);
	}
}


void seq_bender(int dev, int chn, int p1, int p2)
{
	int  i, val;

	DEB1(fprt(stderr, "__GSI: seq_bender dev=%d, chn=%d, p1=%d, p2=%d\n", dev, chn, p1, p2));
	val = (p2 << 7) + p1;
	channel[chn].bender = val;

	switch(synth_types[channel_output[chn]]) {
	case MIDI_HW:
		SEQ_MIDIOUT(dev, MIDI_PITCH_BEND + chn);
		SEQ_MIDIOUT(dev, p1);
		SEQ_MIDIOUT(dev, p2);
		break;
	case SYNTH_HW_AWE32:
		SEQ_BENDER(dev, chn, val);
		break;
	default:
		for (i = 0; i < card_info[dev].nr_voices; i++)
			if (voices[dev][i].channel == chn)
				SEQ_BENDER(dev, i, val);
	}
}


/*---------------------------------------------------------------------------*
 * pause and continue don't work properly yet, MIDI timer doesn't seem to
 * stop
 *---------------------------------------------------------------------------*/
void seq_pause () {
	int  i;
	int  dev;

//	ioctl(seqfd, SNDCTL_SEQ_RESET);

	for (dev=0; dev<nrsynths; dev++) {
		for (i = 0; i < card_info[dev].nr_voices; i++)
#if 1
			SEQ_STOP_NOTE(dev, i, voices[dev][i].note, 64);
#else
			if (voices[dev][i].channel == chn && voices[dev][i].note == note) {
				voices[dev][i].available = 1;
				voices[dev][i].timestamp /= 2;
				if (!channel[chn].controller[CTL_SUSTAIN] && !IS_PERC(chn))
					SEQ_STOP_NOTE(dev, i, note, vel);
			}
#endif
	}
	while(_seqbufptr) { SEQ_DUMPBUF(); }
	SEQ_STOP_TIMER();
}
void seq_continue () {
	int  i;
	int  dev;

	for (dev=0; dev<nrsynths; dev++) {
		for (i = 0; i < card_info[dev].nr_voices; i++)
#if 1
//			SEQ_STOP_NOTE(dev, i, voices[dev][i].note, 10);
;
#else
			if (voices[dev][i].channel == chn && voices[dev][i].note == note) {
				voices[dev][i].available = 1;
				voices[dev][i].timestamp /= 2;
				if (!channel[chn].controller[CTL_SUSTAIN] && !IS_PERC(chn))
					SEQ_STOP_NOTE(dev, i, note, vel);
			}
#endif
	}
	while(_seqbufptr) { SEQ_DUMPBUF(); }
	SEQ_CONTINUE_TIMER();
}

#endif /* USE_OSS */

/*---------------------------------------------------------------------------*
 *	EOF gsi_midi_synth_control.c
 *---------------------------------------------------------------------------*/
